@@ -32,146 +32,123 @@ struct rpmds_s {
     int32_t Count;		/*!< No. of elements */
     unsigned int instance;	/*!< From rpmdb instance? */
     int i;			/*!< Element index. */
-    int nopromote;		/*!< Don't promote Epoch: in rpmdsCompare()? */
     int nrefs;			/*!< Reference count. */
     int *ti;			/*!< Trigger index. */
 };
 
+struct depinfo_s {
+    rpmTagVal typeTag;
+    rpmTagVal evrTag;
+    rpmTagVal flagTag;
+    rpmTagVal ixTag;
+    const char *name;
+    char abrev;
+};
+
+static const struct depinfo_s depTypes[] = {
+    {	RPMTAG_PROVIDENAME,		RPMTAG_PROVIDEVERSION,
+	RPMTAG_PROVIDEFLAGS,		0,
+	"Provides",			'P',
+    },
+    {	RPMTAG_REQUIRENAME,		RPMTAG_REQUIREVERSION,
+	RPMTAG_REQUIREFLAGS,		0,
+	"Requires",			'R',
+    },
+    {	RPMTAG_CONFLICTNAME,		RPMTAG_CONFLICTVERSION,
+	RPMTAG_CONFLICTFLAGS,		0,
+	"Conflicts",			'C',
+    },
+    {	RPMTAG_OBSOLETENAME,		RPMTAG_OBSOLETEVERSION,
+	RPMTAG_OBSOLETEFLAGS,		0,
+	"Obsoletes",			'O',
+    },
+    {	RPMTAG_SUPPLEMENTNAME,		RPMTAG_SUPPLEMENTVERSION,
+	RPMTAG_SUPPLEMENTFLAGS,		0,
+	"Supplements",			'S',
+    },
+    { 	RPMTAG_ENHANCENAME,		RPMTAG_ENHANCEVERSION,
+	RPMTAG_ENHANCEFLAGS,		0,
+	"Enhances",			'e',
+    },
+    {	RPMTAG_RECOMMENDNAME,		RPMTAG_RECOMMENDVERSION,
+	RPMTAG_RECOMMENDFLAGS,		0,
+	"Recommends",			'r',
+    },
+    {	RPMTAG_SUGGESTNAME,		RPMTAG_SUGGESTVERSION,
+	RPMTAG_SUGGESTFLAGS,		0,
+	"Suggests",			's',
+    },
+    {	RPMTAG_ORDERNAME,		RPMTAG_ORDERVERSION,
+	RPMTAG_ORDERFLAGS,		0,
+	"Order",			'o',
+    },
+    {	RPMTAG_TRIGGERNAME,		RPMTAG_TRIGGERVERSION,
+	RPMTAG_TRIGGERFLAGS,		RPMTAG_TRIGGERINDEX,
+	"Trigger",			't',
+    },
+    {	RPMTAG_FILETRIGGERNAME,		RPMTAG_FILETRIGGERVERSION,
+	RPMTAG_FILETRIGGERFLAGS,	RPMTAG_FILETRIGGERINDEX,
+	"FileTrigger",			'f',
+    },
+    {	RPMTAG_TRANSFILETRIGGERNAME,	RPMTAG_TRANSFILETRIGGERVERSION,
+	RPMTAG_TRANSFILETRIGGERFLAGS,	RPMTAG_TRANSFILETRIGGERINDEX,
+	"TransFileTrigger",		'F',
+    },
+    {	RPMTAG_OLDSUGGESTSNAME,		RPMTAG_OLDSUGGESTSVERSION,
+	RPMTAG_OLDSUGGESTSFLAGS,	0,
+	"Oldsuggests",			'?',
+    },
+    {	RPMTAG_OLDENHANCESNAME,		RPMTAG_OLDENHANCESVERSION,
+	RPMTAG_OLDENHANCESFLAGS,	0,
+	"Oldenhances",			'?',
+    },
+    {	0,				0,
+	0,				0,
+	NULL,				0,
+    }
+};
+
+static const struct depinfo_s *depinfoByTag(rpmTagVal tag)
+{
+    for (const struct depinfo_s *dse = depTypes; dse->name; dse++) {
+	if (tag == dse->typeTag)
+	    return dse;
+    }
+    return NULL;
+}
+
+static const struct depinfo_s *depinfoByAbrev(char abrev)
+{
+    for (const struct depinfo_s *dse = depTypes; dse->name; dse++) {
+	if (abrev == dse->abrev)
+	    return dse;
+    }
+    return NULL;
+}
 static int dsType(rpmTagVal tag, 
 		  const char ** Type, rpmTagVal * tagEVR, rpmTagVal * tagF,
 		  rpmTagVal * tagTi)
 {
-    int rc = 0;
-    const char *t = NULL;
-    rpmTagVal evr = RPMTAG_NOT_FOUND;
-    rpmTagVal f = RPMTAG_NOT_FOUND;
-    rpmTagVal ti = RPMTAG_NOT_FOUND;
-
-    if (tag == RPMTAG_PROVIDENAME) {
-	t = "Provides";
-	evr = RPMTAG_PROVIDEVERSION;
-	f = RPMTAG_PROVIDEFLAGS;
-    } else if (tag == RPMTAG_REQUIRENAME) {
-	t = "Requires";
-	evr = RPMTAG_REQUIREVERSION;
-	f = RPMTAG_REQUIREFLAGS;
-    } else if (tag == RPMTAG_SUPPLEMENTNAME) {
-	t = "Supplements";
-	evr = RPMTAG_SUPPLEMENTVERSION;
-	f = RPMTAG_SUPPLEMENTFLAGS;
-    } else if (tag == RPMTAG_ENHANCENAME) {
-	t = "Enhances";
-	evr = RPMTAG_ENHANCEVERSION;
-	f = RPMTAG_ENHANCEFLAGS;
-    } else if (tag == RPMTAG_RECOMMENDNAME) {
-	t = "Recommends";
-	evr = RPMTAG_RECOMMENDVERSION;
-	f = RPMTAG_RECOMMENDFLAGS;
-    } else if (tag == RPMTAG_SUGGESTNAME) {
-	t = "Suggests";
-	evr = RPMTAG_SUGGESTVERSION;
-	f = RPMTAG_SUGGESTFLAGS;
-    } else if (tag == RPMTAG_CONFLICTNAME) {
-	t = "Conflicts";
-	evr = RPMTAG_CONFLICTVERSION;
-	f = RPMTAG_CONFLICTFLAGS;
-    } else if (tag == RPMTAG_OBSOLETENAME) {
-	t = "Obsoletes";
-	evr = RPMTAG_OBSOLETEVERSION;
-	f = RPMTAG_OBSOLETEFLAGS;
-    } else if (tag == RPMTAG_ORDERNAME) {
-	t = "Order";
-	evr = RPMTAG_ORDERVERSION;
-	f = RPMTAG_ORDERFLAGS;
-    } else if (tag == RPMTAG_TRIGGERNAME) {
-	t = "Trigger";
-	evr = RPMTAG_TRIGGERVERSION;
-	f = RPMTAG_TRIGGERFLAGS;
-	ti = RPMTAG_TRIGGERINDEX;
-    } else if (tag == RPMTAG_OLDSUGGESTSNAME) {
-	t = "Oldsuggests";
-	evr = RPMTAG_OLDSUGGESTSVERSION;
-	f = RPMTAG_OLDSUGGESTSFLAGS;
-    } else if (tag == RPMTAG_OLDENHANCESNAME) {
-	t = "Oldenhances";
-	evr = RPMTAG_OLDENHANCESVERSION;
-	f = RPMTAG_OLDENHANCESFLAGS;
-    } else {
-	rc = 1;
+    const struct depinfo_s *di = depinfoByTag(tag);
+    if (di) {
+	if (Type) *Type = di->name;
+	if (tagEVR) *tagEVR = di->evrTag;
+	if (tagF) *tagF = di->flagTag;
+	if (tagTi) *tagTi = di->ixTag;
     } 
-    if (Type) *Type = t;
-    if (tagEVR) *tagEVR = evr;
-    if (tagF) *tagF = f;
-    if (tagTi) *tagTi = ti;
-    return rc;
+    return (di == NULL);
 }    
 
 static char tagNToChar(rpmTagVal tagN)
 {
-    switch (tagN) {
-    default:
-	return 'R';
-	break;
-    case RPMTAG_REQUIRENAME:
-	return 'R';
-	break;
-    case RPMTAG_PROVIDENAME:
-	return 'P';
-	break;
-    case RPMTAG_RECOMMENDNAME:
-	return 'r';
-	break;
-    case RPMTAG_SUGGESTNAME:
-	return 's';
-	break;
-    case RPMTAG_SUPPLEMENTNAME:
-	return 'S';
-	break;
-    case RPMTAG_ENHANCENAME:
-	return 'e';
-	break;
-    case RPMTAG_CONFLICTNAME:
-	return 'C';
-	break;
-    case RPMTAG_OBSOLETENAME:
-	return 'O';
-	break;
-    }
+    const struct depinfo_s *di = depinfoByTag(tagN);
+    return (di != NULL) ? di->abrev : '\0';
 }
 
 rpmTagVal rpmdsDToTagN(char deptype)
 {
-    rpmTagVal tagN = RPMTAG_REQUIRENAME;
-    switch (deptype) {
-    default:
-	tagN = RPMTAG_NOT_FOUND;
-	break;
-    case 'P':
-	tagN = RPMTAG_PROVIDENAME;
-	break;
-    case 'R':
-	tagN = RPMTAG_REQUIRENAME;
-	break;
-    case 'r':
-	tagN = RPMTAG_RECOMMENDNAME;
-	break;
-    case 's':
-	tagN = RPMTAG_SUGGESTNAME;
-	break;
-    case 'S':
-	tagN = RPMTAG_SUPPLEMENTNAME;
-	break;
-    case 'e':
-	tagN = RPMTAG_ENHANCENAME;
-	break;
-    case 'C':
-	tagN = RPMTAG_CONFLICTNAME;
-	break;
-    case 'O':
-	tagN = RPMTAG_OBSOLETENAME;
-	break;
-    }
-    return tagN;
+    const struct depinfo_s *di = depinfoByAbrev(deptype);
+    return (di != NULL) ? di->typeTag : RPMTAG_NOT_FOUND;
 }
 
 rpmsid rpmdsNIdIndex(rpmds ds, int i)
@@ -283,7 +260,6 @@ static rpmds rpmdsCreate(rpmstrPool pool,
     ds->Type = Type;
     ds->Count = Count;
     ds->instance = instance;
-    ds->nopromote = _rpmds_nopromote;
     ds->i = -1;
 
     return rpmdsLink(ds);
@@ -299,20 +275,38 @@ rpmds rpmdsNewPool(rpmstrPool pool, Header h, rpmTagVal tagN, int flags)
 	goto exit;
 
     if (headerGet(h, tagN, &names, HEADERGET_MINMEM)) {
-	struct rpmtd_s evr, flags, tindices;
+	struct rpmtd_s evr, dflags, tindices;
+	rpm_count_t count = rpmtdCount(&names);
 
-	ds = rpmdsCreate(pool, tagN, Type,
-			 rpmtdCount(&names), headerGetInstance(h));
-
-	ds->N = rpmtdToPool(&names, ds->pool);
 	headerGet(h, tagEVR, &evr, HEADERGET_MINMEM);
-	ds->EVR = rpmtdToPool(&evr, ds->pool);
-	headerGet(h, tagF, &flags, HEADERGET_ALLOC);
-	ds->Flags = flags.data;
+	if (evr.count && evr.count != count) {
+	    rpmtdFreeData(&evr);
+	    return NULL;
+	}
+
+	headerGet(h, tagF, &dflags, HEADERGET_ALLOC);
+	if (dflags.count && dflags.count != count) {
+	    rpmtdFreeData(&dflags);
+	    return NULL;
+	}
+
 	if (tagTi != RPMTAG_NOT_FOUND) {
 	    headerGet(h, tagTi, &tindices, HEADERGET_ALLOC);
+	    if (tindices.count && tindices.count != count) {
+		rpmtdFreeData(&tindices);
+		return NULL;
+	    }
+	}
+
+	ds = rpmdsCreate(pool, tagN, Type, count, headerGetInstance(h));
+
+	ds->N = names.count ? rpmtdToPool(&names, ds->pool) : NULL;
+	ds->EVR = evr.count ? rpmtdToPool(&evr, ds->pool): NULL;
+	ds->Flags = dflags.data;
+	if (tagTi != RPMTAG_NOT_FOUND) {
 	    ds->ti = tindices.data;
 	}
+
 	/* ensure rpmlib() requires always have RPMSENSE_RPMLIB flag set */
 	if (tagN == RPMTAG_REQUIRENAME && ds->Flags) {
 	    for (int i = 0; i < ds->Count; i++) {
@@ -484,6 +478,44 @@ rpmds rpmdsCurrent(rpmds ds)
     return cds;
 }
 
+rpmds rpmdsFilterTi(rpmds ds, int ti)
+{
+    int i, i2, tiCount = 0;
+    rpmds fds;
+
+    if (ds == NULL || !ds->ti || !ds->Count)
+	return NULL;
+
+    for (i = 0; i < ds->Count; i++) {
+	if (ds->ti[i] == ti)
+	    tiCount++;
+    }
+
+    if (!tiCount)
+	return NULL;
+
+    fds = rpmdsCreate(ds->pool, ds->tagN, ds->Type, tiCount, ds->instance);
+
+    fds->N = xmalloc(tiCount * sizeof(*fds->N));
+    fds->EVR = xmalloc(tiCount * sizeof(*fds->EVR));
+    fds->Flags = xmalloc(tiCount * sizeof(*fds->Flags));
+    fds->ti = xmalloc(tiCount * sizeof(*fds->ti));
+    fds->i = -1;
+
+    i2 = 0;
+    for (i = 0; i < ds->Count; i++) {
+	if (ds->ti[i] == ti) {
+	    fds->N[i2] = ds->N[i];
+	    fds->EVR[i2] = ds->EVR[i];
+	    fds->Flags[i2] = ds->Flags[i];
+	    fds->ti[i2] = ds->ti[i];
+	    i2++;
+	}
+    }
+
+    return fds;
+}
+
 int rpmdsPutToHeader(rpmds ds, Header h)
 {
     rpmTagVal tagN = rpmdsTagN(ds);
@@ -625,22 +657,12 @@ unsigned int rpmdsInstance(rpmds ds)
 
 int rpmdsNoPromote(const rpmds ds)
 {
-    int nopromote = 0;
-
-    if (ds != NULL)
-	nopromote = ds->nopromote;
-    return nopromote;
+    return 1;
 }
 
 int rpmdsSetNoPromote(rpmds ds, int nopromote)
 {
-    int onopromote = 0;
-
-    if (ds != NULL) {
-	onopromote = ds->nopromote;
-	ds->nopromote = nopromote;
-    }
-    return onopromote;
+    return 1;
 }
 
 rpm_color_t rpmdsColor(const rpmds ds)
@@ -714,7 +736,6 @@ static rpmds rpmdsDup(const rpmds ods)
     size_t nb;
     
     ds->i = ods->i;
-    ds->nopromote = ods->nopromote;
 
     nb = ds->Count * sizeof(*ds->N);
     ds->N = memcpy(xmalloc(nb), ods->N, nb);
@@ -939,108 +960,6 @@ int rpmdsSearch(rpmds ds, rpmds ods)
     }
     return i;
 }
-/**
- * Split EVR into epoch, version, and release components.
- * @param evr		[epoch:]version[-release] string
- * @retval *ep		pointer to epoch
- * @retval *vp		pointer to version
- * @retval *rp		pointer to release
- */
-static
-void parseEVR(char * evr,
-		const char ** ep,
-		const char ** vp,
-		const char ** rp)
-{
-    const char *epoch;
-    const char *version;		/* assume only version is present */
-    const char *release;
-    char *s, *se;
-
-    s = evr;
-    while (*s && risdigit(*s)) s++;	/* s points to epoch terminator */
-    se = strrchr(s, '-');		/* se points to version terminator */
-
-    if (*s == ':') {
-	epoch = evr;
-	*s++ = '\0';
-	version = s;
-	if (*epoch == '\0') epoch = "0";
-    } else {
-	epoch = NULL;	/* XXX disable epoch compare if missing */
-	version = evr;
-    }
-    if (se) {
-	*se++ = '\0';
-	release = se;
-    } else {
-	release = NULL;
-    }
-
-    if (ep) *ep = epoch;
-    if (vp) *vp = version;
-    if (rp) *rp = release;
-}
-
-static inline int rpmdsCompareEVR(const char *AEVR, uint32_t AFlags,
-				  const char *BEVR, uint32_t BFlags,
-				  int nopromote)
-{
-    const char *aE, *aV, *aR, *bE, *bV, *bR;
-    char *aEVR = xstrdup(AEVR);
-    char *bEVR = xstrdup(BEVR);
-    int sense = 0;
-    int result = 0;
-
-    parseEVR(aEVR, &aE, &aV, &aR);
-    parseEVR(bEVR, &bE, &bV, &bR);
-
-    /* Compare {A,B} [epoch:]version[-release] */
-    if (aE && *aE && bE && *bE)
-	sense = rpmvercmp(aE, bE);
-    else if (aE && *aE && atol(aE) > 0) {
-	if (!nopromote) {
-	    sense = 0;
-	} else
-	    sense = 1;
-    } else if (bE && *bE && atol(bE) > 0)
-	sense = -1;
-
-    if (sense == 0) {
-	sense = rpmvercmp(aV, bV);
-	if (sense == 0) {
-	    if (aR && *aR && bR && *bR) {
-		sense = rpmvercmp(aR, bR);
-	    } else {
-		/* always matches if the side with no release has SENSE_EQUAL */
-		if ((aR && *aR && (BFlags & RPMSENSE_EQUAL)) ||
-		    (bR && *bR && (AFlags & RPMSENSE_EQUAL))) {
-		    aEVR = _free(aEVR);
-		    bEVR = _free(bEVR);
-		    result = 1;
-		    goto exit;
-		}
-	    }
-	}
-    }
-
-    /* Detect overlap of {A,B} range. */
-    if (sense < 0 && ((AFlags & RPMSENSE_GREATER) || (BFlags & RPMSENSE_LESS))) {
-	result = 1;
-    } else if (sense > 0 && ((AFlags & RPMSENSE_LESS) || (BFlags & RPMSENSE_GREATER))) {
-	result = 1;
-    } else if (sense == 0 &&
-	(((AFlags & RPMSENSE_EQUAL) && (BFlags & RPMSENSE_EQUAL)) ||
-	 ((AFlags & RPMSENSE_LESS) && (BFlags & RPMSENSE_LESS)) ||
-	 ((AFlags & RPMSENSE_GREATER) && (BFlags & RPMSENSE_GREATER)))) {
-	result = 1;
-    }
-
-exit:
-    free(aEVR);
-    free(bEVR);
-    return result;
-}
 
 int rpmdsCompareIndex(rpmds A, int aix, rpmds B, int bix)
 {
@@ -1076,7 +995,13 @@ int rpmdsCompareIndex(rpmds A, int aix, rpmds B, int bix)
 	result = 1;
     } else {
 	/* Both AEVR and BEVR exist, compare [epoch:]version[-release]. */
-	result = rpmdsCompareEVR(AEVR, AFlags, BEVR, BFlags, B->nopromote);
+	rpmver av = rpmverParse(AEVR);
+	rpmver bv = rpmverParse(BEVR);
+
+	result = rpmverOverlap(av, AFlags, bv, BFlags);
+
+	rpmverFree(av);
+	rpmverFree(bv);
     }
 
 exit:
@@ -1089,7 +1014,7 @@ int rpmdsCompare(const rpmds A, const rpmds B)
 }
 
 int rpmdsMatches(rpmstrPool pool, Header h, int prix,
-		 rpmds req, int selfevr, int nopromote)
+		 rpmds req, int selfevr)
 {
     rpmds provides;
     rpmTagVal tag = RPMTAG_PROVIDENAME;
@@ -1101,8 +1026,6 @@ int rpmdsMatches(rpmstrPool pool, Header h, int prix,
     else
 	provides = rpmdsNewPool(pool, h, tag, 0);
 
-    rpmdsSetNoPromote(provides, nopromote);
-
     /*
      * For a self-provide and indexed provide, we only need one comparison.
      * Otherwise loop through the provides until match or end.
@@ -1127,17 +1050,17 @@ int rpmdsMatches(rpmstrPool pool, Header h, int prix,
 
 int rpmdsMatchesDep (const Header h, int ix, const rpmds req, int nopromote)
 {
-    return rpmdsMatches(NULL, h, ix, req, 0, nopromote);
+    return rpmdsMatches(NULL, h, ix, req, 0);
 }
 
 int rpmdsAnyMatchesDep (const Header h, const rpmds req, int nopromote)
 {
-    return rpmdsMatches(NULL, h, -1, req, 0, nopromote);
+    return rpmdsMatches(NULL, h, -1, req, 0);
 }
 
 int rpmdsNVRMatchesDep(const Header h, const rpmds req, int nopromote)
 {
-    return rpmdsMatches(NULL, h, -1, req, 1, nopromote);
+    return rpmdsMatches(NULL, h, -1, req, 1);
 }
 
 /**
@@ -1187,11 +1110,9 @@ static const struct rpmlibProvides_s rpmlibProvides[] = {
     { "rpmlib(ConcurrentAccess)",    "4.1-1",
 	(                RPMSENSE_EQUAL),
     N_("package scriptlets may access the rpm database while installing.") },
-#ifdef WITH_LUA
     { "rpmlib(BuiltinLuaScripts)",    "4.2.2-1",
 	(                RPMSENSE_EQUAL),
     N_("internal support for lua scripts.") },
-#endif
     { "rpmlib(FileDigests)", 		"4.6.0-1",
 	(		 RPMSENSE_EQUAL),
     N_("file digest algorithm is per package configurable") },
@@ -1206,9 +1127,23 @@ static const struct rpmlibProvides_s rpmlibProvides[] = {
     { "rpmlib(TildeInVersions)",    "4.10.0-1",
 	(		RPMSENSE_EQUAL),
     N_("dependency comparison supports versions with tilde.") },
+    { "rpmlib(CaretInVersions)",    "4.15.0-1",
+	(		RPMSENSE_EQUAL),
+    N_("dependency comparison supports versions with caret.") },
     { "rpmlib(LargeFiles)", 	"4.12.0-1",
 	(		RPMSENSE_EQUAL),
     N_("support files larger than 4GB") },
+    { "rpmlib(RichDependencies)",    "4.12.0-1",
+	(		RPMSENSE_EQUAL),
+    N_("support for rich dependencies.") },
+    { "rpmlib(DynamicBuildRequires)", "4.15.0-1",
+	(RPMSENSE_RPMLIB|RPMSENSE_EQUAL),
+    N_("support for dynamic buildrequires.") },
+#ifdef HAVE_ZSTD
+    { "rpmlib(PayloadIsZstd)",		"5.4.18-1",
+	(RPMSENSE_RPMLIB|RPMSENSE_EQUAL),
+    N_("package payload can be compressed using zstd.") },
+#endif
     { NULL,				NULL, 0,	NULL }
 };
 
@@ -1244,7 +1179,34 @@ rpmstrPool rpmdsPool(rpmds ds)
     return (ds != NULL) ? ds->pool : NULL;
 }
 
-rpmFlags rpmSanitizeDSFlags(rpmTagVal tagN, rpmFlags Flags)
+int rpmdsIsWeak(rpmds ds)
+{
+    int weak = 1;
+    switch (rpmdsTagN(ds)) {
+    case RPMTAG_REQUIRENAME:
+    case RPMTAG_PROVIDENAME:
+    case RPMTAG_OBSOLETENAME:
+    case RPMTAG_CONFLICTNAME:
+	if (!(rpmdsFlags(ds) & RPMSENSE_MISSINGOK))
+	    weak = 0;
+	break;
+    }
+    return weak;
+}
+
+int rpmdsIsReverse(rpmds ds)
+{
+    int reverse = 0;
+    switch (rpmdsTagN(ds)) {
+    case RPMTAG_SUPPLEMENTNAME:
+    case RPMTAG_ENHANCENAME:
+	reverse = 1;
+	break;
+    }
+    return reverse;
+}
+
+rpmsenseFlags rpmSanitizeDSFlags(rpmTagVal tagN, rpmsenseFlags Flags)
 {
     rpmsenseFlags extra = RPMSENSE_ANY;
     switch (tagN) {
@@ -1252,6 +1214,8 @@ rpmFlags rpmSanitizeDSFlags(rpmTagVal tagN, rpmFlags Flags)
 	extra = Flags & RPMSENSE_FIND_PROVIDES;
 	break;
     case RPMTAG_TRIGGERNAME:
+    case RPMTAG_FILETRIGGERNAME:
+    case RPMTAG_TRANSFILETRIGGERNAME:
 	extra = Flags & RPMSENSE_TRIGGER;
 	break;
     case RPMTAG_RECOMMENDNAME:
@@ -1259,10 +1223,375 @@ rpmFlags rpmSanitizeDSFlags(rpmTagVal tagN, rpmFlags Flags)
     case RPMTAG_SUPPLEMENTNAME:
     case RPMTAG_ENHANCENAME:
     case RPMTAG_REQUIRENAME:
-	extra = Flags & _ALL_REQUIRES_MASK;
+    case RPMTAG_ORDERNAME:
+	extra = Flags & (_ALL_REQUIRES_MASK);
+	break;
+    case RPMTAG_CONFLICTNAME:
+	extra = Flags;
 	break;
     default:
 	break;
     }
     return (Flags & RPMSENSE_SENSEMASK) | extra;
 }
+
+static struct ReqComp {
+const char * token;
+    rpmsenseFlags sense;
+} const ReqComparisons[] = { 
+    { "<=", RPMSENSE_LESS | RPMSENSE_EQUAL},
+    { "=<", RPMSENSE_LESS | RPMSENSE_EQUAL},
+    { "<",  RPMSENSE_LESS},
+
+    { "==", RPMSENSE_EQUAL},
+    { "=",  RPMSENSE_EQUAL},
+    
+    { ">=", RPMSENSE_GREATER | RPMSENSE_EQUAL},
+    { "=>", RPMSENSE_GREATER | RPMSENSE_EQUAL},
+    { ">",  RPMSENSE_GREATER},
+
+    { NULL, 0 },
+};
+
+rpmsenseFlags rpmParseDSFlags(const char *str, size_t len)
+{
+    const struct ReqComp *rc;
+    for (rc = ReqComparisons; rc->token != NULL; rc++)
+	if (len == strlen(rc->token) && rstreqn(str, rc->token, len))
+	    return rc->sense;
+    return 0;
+}
+
+static struct RichOpComp {
+    const char * token;
+    rpmrichOp op;
+} const RichOps[] = { 
+    { "and",	 RPMRICHOP_AND},
+    { "or",	 RPMRICHOP_OR},
+    { "if",	 RPMRICHOP_IF},
+    { "unless",	 RPMRICHOP_UNLESS},
+    { "else",	 RPMRICHOP_ELSE},
+    { "with",	 RPMRICHOP_WITH},
+    { "without", RPMRICHOP_WITHOUT},
+    { NULL, 0 },
+};
+
+int rpmdsIsRich(rpmds dep)
+{
+    const char * n = rpmdsN(dep);
+    return (n && n[0] == '(');
+}
+
+static rpmRC parseRichDepOp(const char **dstrp, rpmrichOp *opp, char **emsg)
+{
+    const char *p = *dstrp, *pe = p;
+    const struct RichOpComp *ro;
+
+    while (*pe && !risspace(*pe) && *pe != ')')
+	pe++;
+    for (ro = RichOps; ro->token != NULL; ro++)
+	if (pe - p == strlen(ro->token) && rstreqn(p, ro->token, pe - p)) {
+	    *opp = ro->op;
+	    *dstrp = pe; 
+	    return RPMRC_OK;
+	}
+    if (emsg)
+	rasprintf(emsg, _("Unknown rich dependency op '%.*s'"), (int)(pe - p), p); 
+    return RPMRC_FAIL;
+}
+
+const char *rpmrichOpStr(rpmrichOp op)
+{
+    if (op == RPMRICHOP_SINGLE)
+	return "SINGLE";
+    if (op == RPMRICHOP_AND)
+	return "and";
+    if (op == RPMRICHOP_OR)
+	return "or";
+    if (op == RPMRICHOP_IF)
+	return "if";
+    if (op == RPMRICHOP_UNLESS)
+	return "unless";
+    if (op == RPMRICHOP_ELSE)
+	return "else";
+    if (op == RPMRICHOP_WITH)
+	return "with";
+    if (op == RPMRICHOP_WITHOUT)
+	return "without";
+    return NULL;
+}
+
+
+#define SKIPWHITE(_x)   {while (*(_x) && (risspace(*_x) || *(_x) == ',')) (_x)++;}
+#define SKIPNONWHITEX(_x){int bl = 0; while (*(_x) &&!(risspace(*_x) || *(_x) == ',' || (*(_x) == ')' && bl-- <= 0))) if (*(_x)++ == '(') bl++;}
+
+static rpmRC parseSimpleDep(const char **dstrp, char **emsg, rpmrichParseFunction cb, void *cbdata)
+{
+    const char *p = *dstrp;
+    const char *n, *e = 0;
+    int nl, el = 0;
+    rpmsenseFlags sense = 0;
+
+    n = p;
+    SKIPNONWHITEX(p);
+    nl = p - n;
+    if (nl == 0) {
+        if (emsg)
+          rasprintf(emsg, _("Name required"));
+        return RPMRC_FAIL;
+    }   
+    SKIPWHITE(p);
+    if (*p) {
+        const char *pe = p;
+
+        SKIPNONWHITEX(pe);
+	sense = rpmParseDSFlags(p, pe - p);
+        if (sense) {
+            p = pe; 
+            SKIPWHITE(p);
+            e = p;
+            SKIPNONWHITEX(p);
+            el = p - e;
+        }
+    }   
+    if (e && el == 0) {
+        if (emsg)
+          rasprintf(emsg, _("Version required"));
+        return RPMRC_FAIL;
+    }
+    if (cb && cb(cbdata, RPMRICH_PARSE_SIMPLE, n, nl, e, el, sense, RPMRICHOP_SINGLE, emsg) != RPMRC_OK)
+	return RPMRC_FAIL;
+    *dstrp = p;
+    return RPMRC_OK;
+}
+
+#define RICHPARSE_CHECK		(1 << 0)
+#define RICHPARSE_NO_WITH	(1 << 1)
+#define RICHPARSE_NO_AND	(1 << 2)
+#define RICHPARSE_NO_OR		(1 << 3)
+
+static rpmRC rpmrichParseCheck(rpmrichOp op, int check, char **emsg)
+{
+    if ((op == RPMRICHOP_WITH || op == RPMRICHOP_WITHOUT) && (check & RICHPARSE_NO_WITH) != 0) {
+	if (emsg)
+	    rasprintf(emsg, _("Illegal ops in with/without"));
+	return RPMRC_FAIL;
+    }
+    if ((check & RICHPARSE_CHECK) == 0)
+	return RPMRC_OK;
+    if ((op == RPMRICHOP_AND || op == RPMRICHOP_IF) && (check & RICHPARSE_NO_AND) != 0) {
+	if (emsg)
+	    rasprintf(emsg, _("Illegal context for 'unless', please use 'or' instead"));
+	return RPMRC_FAIL;
+    }
+    if ((op == RPMRICHOP_OR || op == RPMRICHOP_UNLESS) && (check & RICHPARSE_NO_OR) != 0) {
+	if (emsg)
+	    rasprintf(emsg, _("Illegal context for 'if', please use 'and' instead"));
+	return RPMRC_FAIL;
+    }
+    return RPMRC_OK;
+}
+
+static rpmRC rpmrichParseInternal(const char **dstrp, char **emsg, rpmrichParseFunction cb, void *cbdata, int *checkp)
+{
+    const char *p = *dstrp, *pe;
+    rpmrichOp op = RPMRICHOP_SINGLE, firstop = RPMRICHOP_SINGLE, chainop = 0;
+    int check = checkp ? *checkp : 0;
+
+    if (cb && cb(cbdata, RPMRICH_PARSE_ENTER, p, 0, 0, 0, 0, op, emsg) != RPMRC_OK)
+        return RPMRC_FAIL;
+    if (*p++ != '(') {
+        if (emsg)
+          rasprintf(emsg, _("Rich dependency does not start with '('"));
+        return RPMRC_FAIL;
+    }
+    for (;;) {
+        SKIPWHITE(p);
+        if (*p == ')') {
+            if (emsg) {
+                if (chainop)
+                    rasprintf(emsg, _("Missing argument to rich dependency op"));
+                else
+                    rasprintf(emsg, _("Empty rich dependency"));
+            }
+            return RPMRC_FAIL;
+        }
+	if (*p == '(') {
+	    int subcheck = check & RICHPARSE_CHECK;
+	    if (rpmrichParseInternal(&p, emsg, cb, cbdata, &subcheck) != RPMRC_OK)
+		return RPMRC_FAIL;
+	    if (op == RPMRICHOP_IF || op == RPMRICHOP_UNLESS)
+		subcheck &= ~(RICHPARSE_NO_AND | RICHPARSE_NO_OR);
+	    check |= subcheck;
+	} else {
+            if (parseSimpleDep(&p, emsg, cb, cbdata) != RPMRC_OK)
+                return RPMRC_FAIL;
+        }
+        SKIPWHITE(p);
+        if (!*p) {
+            if (emsg)
+                rasprintf(emsg, _("Unterminated rich dependency: %s"), *dstrp);
+            return RPMRC_FAIL;
+        }
+        if (*p == ')')
+            break;
+        pe = p;
+        if (parseRichDepOp(&pe, &op, emsg) != RPMRC_OK)
+            return RPMRC_FAIL;
+	if (firstop == RPMRICHOP_SINGLE)
+	    firstop = op;
+
+	if (op == RPMRICHOP_ELSE && (chainop == RPMRICHOP_IF || chainop == RPMRICHOP_UNLESS))
+	    chainop = 0;
+        if (chainop && op != chainop) {
+            if (emsg)
+                rasprintf(emsg, _("Cannot chain different ops"));
+            return RPMRC_FAIL;
+        }
+        if (chainop && op != RPMRICHOP_AND && op != RPMRICHOP_OR && op != RPMRICHOP_WITH) {
+            if (emsg)
+                rasprintf(emsg, _("Can only chain and/or/with ops"));
+            return RPMRC_FAIL;
+	}
+        if (cb && cb(cbdata, RPMRICH_PARSE_OP, p, pe - p, 0, 0, 0, op, emsg) != RPMRC_OK)
+            return RPMRC_FAIL;
+        chainop = op;
+        p = pe;
+    }
+
+    /* check for illegal combinations */
+    if (rpmrichParseCheck(firstop, check, emsg) != RPMRC_OK)
+	return RPMRC_FAIL;
+
+    /* update check data */
+    if (firstop == RPMRICHOP_IF)
+	check |= RICHPARSE_NO_OR;
+    if (firstop == RPMRICHOP_UNLESS)
+	check |= RICHPARSE_NO_AND;
+    if (op == RPMRICHOP_AND || op == RPMRICHOP_OR)
+	check &= ~(RICHPARSE_NO_AND | RICHPARSE_NO_OR);
+    if (op != RPMRICHOP_SINGLE && op != RPMRICHOP_WITH && op != RPMRICHOP_WITHOUT && op != RPMRICHOP_OR)
+	check |= RICHPARSE_NO_WITH;
+
+    p++;
+    if (cb && cb(cbdata, RPMRICH_PARSE_LEAVE, *dstrp, p - *dstrp , 0, 0, 0, op, emsg) != RPMRC_OK)
+        return RPMRC_FAIL;
+    *dstrp = p;
+    if (checkp)
+	*checkp |= check;
+    return RPMRC_OK;
+}
+
+rpmRC rpmrichParse(const char **dstrp, char **emsg, rpmrichParseFunction cb, void *cbdata)
+{
+    return rpmrichParseInternal(dstrp, emsg, cb, cbdata, NULL);
+}
+
+rpmRC rpmrichParseForTag(const char **dstrp, char **emsg, rpmrichParseFunction cb, void *cbdata, rpmTagVal tagN)
+{
+    int check = RICHPARSE_CHECK;
+    if (rpmrichParseInternal(dstrp, emsg, cb, cbdata, &check) != RPMRC_OK)
+	return RPMRC_FAIL;
+    switch (tagN) {
+    case RPMTAG_CONFLICTNAME:
+    case RPMTAG_SUPPLEMENTNAME:
+    case RPMTAG_ENHANCENAME:
+	if (rpmrichParseCheck(RPMRICHOP_OR, check, emsg) != RPMRC_OK)
+	    return RPMRC_FAIL;
+	break;
+    default:
+	if (rpmrichParseCheck(RPMRICHOP_AND, check, emsg) != RPMRC_OK)
+	    return RPMRC_FAIL;
+	break;
+    }
+    return RPMRC_OK;
+}
+
+struct rpmdsParseRichDepData {
+    rpmds dep;
+    rpmsenseFlags depflags;
+
+    rpmds leftds;
+    rpmds rightds;
+    rpmrichOp op;
+
+    int depth;
+    const char *rightstart;
+    int dochain;
+};
+
+static rpmRC rpmdsParseRichDepCB(void *cbdata, rpmrichParseType type,
+		const char *n, int nl, const char *e, int el, rpmsenseFlags sense,
+		rpmrichOp op, char **emsg) {
+    struct rpmdsParseRichDepData *data = cbdata;
+    rpmds ds = 0;
+
+    if (type == RPMRICH_PARSE_ENTER)
+	data->depth++;
+    else if (type == RPMRICH_PARSE_LEAVE) {
+	if (--data->depth == 0 && data->dochain && data->rightstart) {
+	    /* chain op hack, construct a sub-ds from the right side of the chain */
+	    char *right = xmalloc(n + nl - data->rightstart + 2);
+	    right[0] = '(';
+	    strncpy(right + 1, data->rightstart, n + nl - data->rightstart);
+	    right[n + nl - data->rightstart + 1] = 0;
+	    data->rightds = rpmdsFree(data->rightds);
+	    ds = singleDS(data->dep->pool, data->dep->tagN, 0, 0, data->depflags, 0, 0, 0);
+	    ds->N[0] = rpmstrPoolId(ds->pool, right, 1);
+	    ds->EVR[0] = rpmstrPoolId(ds->pool, "", 1);
+	    data->rightds = ds;
+	    free(right);
+	}
+    }
+    if (data->depth != 1)
+	return RPMRC_OK;	/* we're only interested in top-level parsing */
+    if ((type == RPMRICH_PARSE_SIMPLE || type == RPMRICH_PARSE_LEAVE) && !data->dochain) {
+	if (type == RPMRICH_PARSE_SIMPLE && data->dep->tagN == RPMTAG_REQUIRENAME && nl > 7 &&
+			 rstreqn(n, "rpmlib(", sizeof("rpmlib(")-1))
+	    sense |= RPMSENSE_RPMLIB;
+	ds = singleDS(data->dep->pool, data->dep->tagN, 0, 0, sense | data->depflags, 0, 0, 0);
+	ds->N[0] = rpmstrPoolIdn(ds->pool, n, nl, 1);
+	ds->EVR[0] = rpmstrPoolIdn(ds->pool, e ? e : "", el, 1);
+	if (!data->leftds)
+	    data->leftds = ds;
+	else {
+	    data->rightds = ds;
+	    data->rightstart = n;
+	}
+    }
+    if (type == RPMRICH_PARSE_OP) {
+	if (data->op != RPMRICHOP_SINGLE)
+	    data->dochain = 1;	/* this is a chained op */
+	else
+	    data->op = op;
+    }
+    return RPMRC_OK;
+}
+
+
+rpmRC rpmdsParseRichDep(rpmds dep, rpmds *leftds, rpmds *rightds, rpmrichOp *op, char **emsg)
+{
+    rpmRC rc;
+    struct rpmdsParseRichDepData data;
+    const char *depstr = rpmdsN(dep);
+    memset(&data, 0, sizeof(data));
+    data.dep = dep;
+    data.op = RPMRICHOP_SINGLE;
+    data.depflags = rpmdsFlags(dep) & ~(RPMSENSE_SENSEMASK | RPMSENSE_MISSINGOK);
+    rc = rpmrichParse(&depstr, emsg, rpmdsParseRichDepCB, &data);
+    if (rc == RPMRC_OK && *depstr) {
+	if (emsg)
+	    rasprintf(emsg, _("Junk after rich dependency"));
+	rc = RPMRC_FAIL;
+    }
+    if (rc != RPMRC_OK) {
+	rpmdsFree(data.leftds);
+	rpmdsFree(data.rightds);
+    } else {
+	*leftds = data.leftds;
+	*rightds = data.rightds;
+	*op = data.op;
+    }
+    return rc;
+}
+
